from basic_model import *
from plotting_linscale import *

#We now want to evaluate the model on a grid of parameters beta and gamma
    # and check whether peakinf is monotonically decreasing in severity-perception
    # time_of_peak is monotonically increasing in severity-perception
    # as well as whether finr is monotonically decreasing in severity-perception
    # finally whether duration is monotonically increasing in severity-perception

    # We will vary beta and gamma in the following ranges:
    # beta: [0.1, 0.9]
    # gamma: [0.1, 0.9]
    # We will vary severity in the following range:
    # severity: [0, 100]

def not_all_finished(model):
    return not all([m["finished"] for m in model])

# Outcomes selectors
def peakinf(models):
    return [m["y"][1].max() for m in models]

def time_of_peak(models):
    return [m["t"][np.argmax(m["y"][1, :])] for m in models]

def finr(models):
    return [m["y"][2, -1] for m in models]

def dur(models):
    return [np.argmax(m['y'][1, :] < m['init_infected']) for m in models]

def welfare(models):
    return [m["y"][3, -1] for m in models]

outcomes_selectors = [peakinf, time_of_peak, finr, dur, welfare]
outcome_names = ["Peak of Infections", "Time of Peak", "Total Recovered", "Duration of Epidemic", "Final Welfare"]

# monotonicity and convexity tests
# elemental tests
def is_constant(x):
    return np.all(np.isclose(np.diff(x),0))

def is_monotone_increasing(x):
    return np.all(np.diff(x) >= 0)

def is_monotone_decreasing(x):
    return np.all(np.diff(x) <= 0)


def is_linear(x):
    return np.all(np.isclose(np.diff(np.diff(x)),0))

def is_convex(x):
    return np.all(np.diff(np.diff(x)) >= 0)

def is_concave(x):
    return np.all(np.diff(np.diff(x)) <= 0)


# complete tests
def monotonicity(outcomes, models):
    if not_all_finished(models):
        return np.nan
    
    outcomes = outcomes(models)
    if is_constant(outcomes):
        return 0
    elif is_monotone_increasing(outcomes):
        return 1
    elif is_monotone_decreasing(outcomes):
        return -1
    else:
        return 2
    

def convexity(outcomes, models):
    if not_all_finished(models):
        return np.nan
    
    outcomes = outcomes(models)
    if is_constant(outcomes):
        return 0
    elif is_convex(outcomes):
        return 1
    elif is_concave(outcomes):
        return -1
    else:
        return 2

color_code = {0 : "blue", 
              1 : "green", 
              -1 : "red", 
              2 : "black", 
              np.nan : "grey"}

monot_code = {0 : "constant", 
              1 : "increasing", 
              -1 : "decreasing", 
              2 : "other", 
              np.nan : "unfinished"}

cnvx_code = {0 : "linear", 
             1 : "convex", 
             -1 : "concave", 
             2 : "other", 
             np.nan : "unfinished"}

# change parameters
def set_parameters(parameters, beta, gamma, severity):
    parameters = copy(parameters)
    parameters["beta"] = beta
    parameters["gamma"] = gamma
    parameters["severity"] = severity
    return parameters

parameters = {"init_infected" : 0.0001,
             "beta" : 0.5,
             "gamma": 1/8,
             "severity" : 100,
             "bias" : 1} 

from itertools import product
import time

betal = 0.1
betau = 1
gammal = 0.1
gammau = 0.9
bgdivisions = 10
severityl = 0
severityu = 100
severitydivisions = 20
duration = 5000

def solve_model(params, betal = 0.1, betau = 1, gammal = 0.1, gammau = 0.9, bgdivisions = 10, severityl=0, severityu=100, severitydivisions = 20, duration = 5000):
    beta_values = np.linspace(betal, betau, bgdivisions)  
    gamma_values = np.linspace(gammal, gammau, bgdivisions)  
    #severity_values = np.logspace(severityl, severityu, severitydivisions) 
    severity_values = np.linspace(severityl, severityu, severitydivisions)

    total_number = len(beta_values) * len(gamma_values) * severitydivisions
    print(f"Models to solve: {total_number}")
    nr_solved = 0
    solved_models = {}

    order = 0
    
    bg = {}

    start = time.time()

    for nb, b in enumerate(beta_values):
        for ng, g in enumerate(gamma_values):

            solved_models[(b,g)] = []
            for s in severity_values:
                params = set_parameters(parameters, b, g, s)
                solved_models[(b,g)].append(evaluate(params, duration))
                nr_solved += 1

            if nr_solved / 10 >= order:
                elapsed_time = time.time() - start
                print(f"Solved {nr_solved} of {total_number} models ({nr_solved/(len(beta_values) * len(gamma_values) * severitydivisions) * 100}%) in {elapsed_time} seconds")
                print(f"Estimated time to finish: {(elapsed_time / nr_solved * (total_number - nr_solved))/3600} hours")
                order += 1
    
    return solved_models

# Save the results to a file
import pickle
save = True
if save:
    models = solve_model(parameters)
    with open('models.pkl', 'wb') as f:
        pickle.dump(models, f)
else:
    # open the file in read and binary mode ('rb')
    with open('models.pkl', 'rb') as f:
        # load the data from the file
        models = pickle.load(f)

# 'models' contains now all the solved models

import matplotlib.pyplot as plt
import numpy as np

def create_matrices(models):
    # sets of parameters that are uniformly increasing or decreasing, convex or concave
    # for each outcome


    fig, axs = plt.subplots(4, figsize=(3, 9))
    for n, (sel, name) in enumerate(zip(outcomes_selectors[:-1], outcome_names[:-1])):
        for (b, g), sms in models.items(): # beta b, gamma g, severity- and solved model sequence
            #msq = [m for (s, m) in sms] #  model sequence
            monot_result = monotonicity(sel, sms)
            axs[n].scatter(b, g, color = color_code[monot_result], label = monot_code[monot_result], s=20)
        axs[n].set_title(name)
        axs[n].set_xlabel(r"Transmission $\beta$")
        axs[n].set_ylabel(r"Recovery $\gamma$")

    fig.suptitle("Monotonicity")
    fig.tight_layout()
    fig.show()
    plt.savefig('monotonicity.pdf')


    fig2, axs2 = plt.subplots(4, figsize=(3, 9))
    for n, (sel, name) in enumerate(zip(outcomes_selectors[:-1], outcome_names[:-1])):
        for (b, g), sms in models.items(): # beta b, gamma g, severity- and solved model sequence
            #msq = [m for (s, m) in sms] #  model sequence
            convx_result = convexity(sel, sms)
            axs2[n].scatter(b,g, color = color_code[convx_result], label = monot_code[convx_result], s=20)
        
        axs2[n].set_title(name)
        axs2[n].set_xlabel(r"Transmission $\beta$")
        axs2[n].set_ylabel(r"Recovery $\gamma$")
        
    fig2.suptitle("Convexity")
    fig2.tight_layout()
    fig2.show()
    plt.savefig('convexity.pdf')
    
create_matrices(models)


from scipy.integrate import cumtrapz

# here severityl and severityu are the lower and upper bounds of the REAL severity range
def monotonicity_volume_welfare(models, severityl=0.1, severityu=100, div = 20):
    monotonicity = {}
    convexity = {}
    solved_welfare = {}

    total_number = len(models) * div

    total_count = 0
    non_count = 0
    inc_count = 0
    dec_count = 0
    cont_count = 0
    els_count = 0

    linear_count = 0
    convex_count = 0
    concav_count = 0
    cvx_else_count = 0

    order = 0

    print(f"Models to solve: {total_number}")

    start = time.time()
    # for given parameters beta and gamma
    for (b, g), sm in models.items():

        # for real severity in the range of severityl to severityu
        for theta_real in np.linspace(severityl, severityu, div):
            # for each model m in sm:
            perc_severity_series = []
            for m in sm: 
                theta_perc = m["severity"] # perceived severity of the disease
                # if model not finished, mark and continue
                if not m["finished"]:
                    solved_welfare |= { (b, g, theta_real, theta_perc) : np.nan}
                    #perc_severity_series.append(np.nan)
                    perc_severity_series = None
                else:
                    # Integrate welfare over time for the given pandemic
                    dW = lambda S, I: (-np.minimum(theta_perc*b*I/2,1)**2 - theta_real*b*(1-np.minimum(theta_perc*b*I/2,1))*I)*S
                    W = cumtrapz(dW(m["y"][0], m["y"][1]), m["t"])

                    # save the solution for the given parameter combination
                    solved_welfare |= { (b, g, theta_real, theta_perc) : W}
                    perc_severity_series.append(W[-1])
            
            # note and count monotonicity volumes
            total_count += 1
            
            if perc_severity_series == None:
                monotonicity |= { (b, g, theta_real) : np.nan}
                non_count += 1
            elif is_constant(perc_severity_series):
                monotonicity |= { (b, g, theta_real) : 0}
                cont_count += 1
            elif is_monotone_increasing(perc_severity_series):
                monotonicity |= { (b, g, theta_real) : 1}
                inc_count += 1
            elif is_monotone_decreasing(perc_severity_series):
                monotonicity |= { (b, g, theta_real) : -1}
                dec_count += 1
            else:
                monotonicity |= { (b, g, theta_real) : 2}
                els_count += 1

            if perc_severity_series == None:
                convexity |= { (b, g, theta_real) : np.nan}
            elif is_linear(perc_severity_series):
                convexity |= { (b, g, theta_real) : 1}
                linear_count += 1
            elif is_convex(perc_severity_series):
                convexity |= { (b, g, theta_real) : 0}
                convex_count += 1
            elif is_concave(perc_severity_series):
                convexity |= { (b, g, theta_real) : -1}
                concav_count += 1
            else:
                convexity |= { (b, g, theta_real) : 2}
                cvx_else_count += 1
            
            if total_count / (total_number*0.01) >= order: #give output every 1 percentage point
                if perc_severity_series != None:
                    fig, axs = plt.subplots()
                    x = [m["severity"] for m in sm]
                    y = perc_severity_series
                    axs.plot(x, y)
                    fig.show()
                else:
                    print("encountered unfinished model")

                elapsed_time = time.time() - start
                print(f"Solved {total_count} of {total_number} models ({total_count/total_number * 100}%) in {elapsed_time} seconds")
                print(f"Estimated time to finish: {(elapsed_time / total_count * (total_number - total_count))/3600} hours")
                order += 1

    
    return total_count, non_count, \
            inc_count, cont_count, dec_count, els_count, \
            linear_count, convex_count, concav_count, cvx_else_count, \
            monotonicity, solved_welfare

total_count, non_count, \
    inc_count, cont_count, dec_count, els_count, \
        linear_count, convex_count, concav_count, cvx_else_count, \
            monotone, solved_welfare = monotonicity_volume_welfare(models)

print(f"solved: {total_count}")
print(f"non finished: {non_count}, {non_count/total_count * 100}%")
print("-------------------")
print(f"increasing: {inc_count}, {inc_count/total_count * 100}%")
print(f"constant: {cont_count}, {cont_count/total_count * 100}%")
print(f"decreasing: {dec_count}, {dec_count/total_count * 100}%")
print(f"else: {els_count}, {els_count/total_count * 100}%")
print("-------------------")
print(f"convex: {dec_count}, {dec_count/total_count * 100}%")
print(f"linear: {inc_count}, {inc_count/total_count * 100}%")
print(f"concav: {cont_count}, {cont_count/total_count * 100}%")
print(f"else: {cvx_else_count}, {els_count/total_count * 100}%")



