from basic_model import *

import matplotlib as mlt
import matplotlib.pyplot as plt
import matplotlib.colors as colors
result_names = {0 : "susceptible",
                1 : "infected",
                2 : "recovered",
                3 : "welfare"}

def plot(evaluated_params, param_name, title_text="", time = -1, highlight = -1):
    """
    evaluated_params - a sequence of evaluated parameters where only param_name varies
    0 -> susceptible
    1 -> infected
    2 -> recovered
    3 -> welfare
    """
    result_nrs = (0, 1, 2, 3)
    fig, ax = plt.subplots(4, 1, figsize=(10, 14))

    # auf ax[i] ergebnis nr i plotten
    rng = [p[param_name] for p in evaluated_params]
    mn = min(rng)
    mx = max(rng)
    norm = mlt.colors.Normalize(mn, mx)
    cmap = mlt.colormaps["plasma"]

    #fig.suptitle(title_text)
    
    for j, p in enumerate(evaluated_params):
        t = p["t"]
        y = p["y"]
        for i, result_nr in enumerate(result_nrs):
            ax[i].set_title(result_names[result_nr])

            if j == highlight:
                ax[i].plot(t, y[result_nr, :], label=f'{param_name}: {p[param_name]:.2f}', color="black") #, final w: {y[3, -1]:.2f}, final rec.: {y[2, -1]:.2f}
            else:
                ax[i].plot(t, y[result_nr, :], label=f'{param_name}: {p[param_name]:.2f}', color=cmap(norm(p[param_name]))) #, final w: {y[3, -1]:.2f}, final rec.: {y[2, -1]:.2f}
    
    #ax[0].legend()
    #plt.subplots_adjust(top=0.95, bottom=0.05)
    #plt.savefig(f'dynamics_{param_name}_{int(evaluated_params[0]["t"][-1])}.png')

    # Create separate plot for the legend
    fig_legend, ax_legend = plt.subplots(figsize=(2, 3.5))
    ax_legend.axis('off')
    handles, labels = ax[0].get_legend_handles_labels()
    ax_legend.legend(handles, labels, loc='center', ncol=1)
    
    #plt.tight_layout()
    #plt.savefig(f'dynamics_legend_{param_name}_{int(evaluated_params[0]["t"][-1])}.png')
    #plt.show()
    

# We are now going to plot how the various epidemiological measures depend on disease perception
# We will vary the severity of the disease and plot:
#   the peak of infections, the time of the peak, final recovered, duration of the epidemic

def plot_peak(evaluated_parameters):
    """
    Plot the peak of infections against the severity of the disease
    0 -> susceptible
    1 -> infected
    2 -> recovered
    3 -> welfare
    """
    fig, ax = plt.subplots(2, 1, figsize=(15, 10))

    #fig.suptitle(f"gamma: {evaluated_parameters[0]['gamma']}, beta: {evaluated_parameters[0]['beta']}")
    
    severity = [m["severity"] for m in evaluated_parameters]

    peakinf = [m["y"][1].max() for m in evaluated_parameters]
    time_of_peak = [np.argmax(m["y"][1, :]) for m in evaluated_parameters]  # Find time of peak infection

    ax[0].plot(severity, peakinf, label=f"peak infections", color="black")
    ax[0].set_title("peak infections")
    ax[0].set_xlabel("severity")
    #ax[0].set_xscale('log')
    #ax[0].set_yscale('symlog')

    ax[1].plot(severity, time_of_peak, label=f"time of peak", color="black")
    ax[1].set_title("time of peak")
    ax[1].set_xlabel("severity")
    #ax[1].set_xscale('log')

    plt.subplots_adjust(top=0.95, bottom=0.05)

    plt.savefig(f'final_infected.png')
    plt.show()
    
def plot_final(evaluated_parameters):
    """
    Plot final recovered and duration of the epidemic

    0 -> susceptible
    1 -> infected
    2 -> recovered
    3 -> welfare
    """
    fig, ax = plt.subplots(2, 1, figsize=(15, 10))

    #fig.suptitle(f"gamma: {evaluated_parameters[0]['gamma']}, beta: {evaluated_parameters[0]['beta']}")
    
    severity = [m["severity"] for m in evaluated_parameters]

    finr = [m["y"][2, -1] for m in evaluated_parameters]

    #find mininum time where the epidemic is finished (as defined above) for every m in evaluated_parameters
    duration = [np.argmax(m['y'][1, :] < m['init_infected']) for m in evaluated_parameters]
    
    ax[0].plot(severity, finr, label=f"final recovered", color="black")

    ax[0].set_title("final recovered")
    ax[0].set_xlabel("severity")
    #ax[0].set_xscale('log')

    ax[1].plot(severity, duration, label=f"duration", color="black")
    ax[1].set_title("duration")
    ax[1].set_xlabel("severity")
    #ax[1].set_xscale('log')
    
    plt.subplots_adjust(top=0.95, bottom=0.05)

    plt.savefig(f'final_recovered_duration.png')
    plt.show()

def plot_welfare(models):
    """
    Plot final welfare against severity
    0 -> susceptible
    1 -> infected
    2 -> recovered
    3 -> welfare
    """
    fig, ax = plt.subplots(1, 1, figsize=(15, 10))

    #fig.suptitle(f"gamma: {models[0]['gamma']}, beta: {models[0]['beta']}")
    
    severity = [m["severity"] for m in models]

    finw = [m["y"][3, -1] for m in models]
    
    ax.plot(severity, finw, label=f"final recovered", color="black")

    ax.set_title("final recovered")
    ax.set_xlabel("severity")
    #ax.set_xscale('log')

    plt.subplots_adjust(top=0.95, bottom=0.05)

    plt.savefig(f'final_welfare.png')
    plt.show()

def plot_welfare_bias(models):
    """
    0 -> susceptible
    1 -> infected
    2 -> recovered
    3 -> welfare
    """

    fig, ax = plt.subplots(1, 1, figsize=(15, 10))

    fig.suptitle(f"gamma: {models[0]['gamma']}, beta: {models[0]['beta']}, severity: {models[0]['severity']}")
    
    bias = [m["bias"] for m in models]

    finw = [m["y"][3, -1] for m in models]
    
    ax.plot(bias, finw, label=f"welfare", color="black")

    ax.set_title("final recovered")
    ax.set_xlabel("bias")
    #ax.set_xscale('log')

    plt.subplots_adjust(top=0.95, bottom=0.05)

    plt.savefig(f'final_welfare_by_bias_.png')
    plt.show()